"""Generic Acquia Cloud API resource."""

from .acquiadata import AcquiaData


class AcquiaResource(AcquiaData):

    """Acquia Cloud API resource."""

    #: Valid properties for this resource object
    valid_keys = None

    def __getitem__(self, key):
        """ Get the value of an object property. """
        if None == self.data:
            self.get()

        if None != self.valid_keys and key in self.valid_keys:
            return self.data[key]

    def get(self):
        """Return the resource from the Acquia Cloud API."""
        if None == self.data:
            response = self.request()
            self.data = response

        return self.data



""" Acquia Cloud API server resource. """

from .acquiaresource import AcquiaResource


class Server(AcquiaResource):

    """Acquia Cloud API Server."""

    valid_keys = ['name', 'fqdn', 'services', 'status', 'ami_type', 'ec2_region',
                  'ec2_availability_zone']




""" Acquia API task queue resource. """

from datetime import datetime
import logging
import re
import requests_cache
import time

from .acquiaresource import AcquiaResource
from ..exceptions import AcquiaCloudTaskFailedException

LOGGER = logging.getLogger('acapi.resources.task')

class Task(AcquiaResource):

    """ Task queue resource. """

    #: Task polling interval in seconds.
    POLL_INTERVAL = 3

    #: Valid task properties
    valid_keys = ['body', 'completed', 'cookie', 'created', 'description', 'hidden',
                  'id', 'percentage', 'queue', 'received', 'recipient', 'result',
                  'sender', 'started', 'state']

    def __init__(self, uri, auth, data=None, hack_uri=True):
        """ Constructor.

        Parameters
        ----------
        uri : str
            The base URI for the resource.
        auth : tuple
             The authentication credentials to use for the request.
        data : dict
            Raw data from ACAPI.
        hack_uri : bool
            Hack the URI so it is valid?
        """
        if hack_uri:
            uri = self.mangle_uri(uri, data)

        self.loops = 0

        AcquiaResource.__init__(self, uri, auth, data)

    def mangle_uri(self, uri, task_data):
        """Generate a URI for a task based on JSON task object.

        Parameters
        ----------
        task_data : dict
            Raw task data from ACAPI.

        Returns
        -------
        Task
            Task object.
        """
        task_id = int(task_data['id'])

        pattern = re.compile(r'/sites/([a-z\:0-9]+)(/.*)?')
        task_uri = pattern.sub(r'/sites/\g<1>/tasks/{}'.format(task_id), uri)

        return task_uri

    def pending(self):
        """ Check if a task is still pending.

        Returns
        -------
        bool
            Is the task still pending completion?
        """
        # Ensure we don't have stale data
        
        # Disable caching so we get the real response
        with requests_cache.disabled():
            task = self.request()

        self.data = task
        return None == task['completed']

    def wait(self):
        """Wait for a task to finish executing.

        Returns
        -------
        Task
            Task object.
        """
        # We wait a maximum of 30mins
        max_loops = (60 * 30 / self.POLL_INTERVAL)
        start = datetime.now()

        while self.pending():
            if self.loops >= max_loops:
                break

            self.loops += 1
            time.sleep(self.POLL_INTERVAL)

        # Grab the cached response
        task = self.get()
        if 'done' != task['state']:
            raise AcquiaCloudTaskFailedException('Task {task_id} failed'
                                                 .format(task_id=task['id']), task)

        end = datetime.now()
        delta = end - start
        LOGGER.info("Waited %.2f seconds for task to complete", delta.total_seconds())

        return self



""" Acquia Cloud API Environment resource. """

import re
from .acquiaresource import AcquiaResource
from .database import Database
from .databaselist import DatabaseList
from .domain import Domain
from .domainlist import DomainList
from .server import Server
from .serverlist import ServerList

class Environment(AcquiaResource):

    """Environment associated with a site."""

    #: Valid keys for environment object.
    valid_keys = ['name', 'vcs_path', 'ssh_host', 'db_clusters', 'default_domain', 'livedev']

    def copy_files(self, target):
        """Copy files to another environment.

        Parameters
        ----------
        target : str
            The name of the environment to copy the files to.

        Returns
        -------
        bool
            Were the files successfully copied?
        """
        pattern = re.compile('/envs/(.*)')
        base_uri = pattern.sub(r'/files-copy/\g<1>', self.uri)
        uri = '{uri}/{target}'.format(uri=base_uri, target=target)

        task_data = self.request(uri=uri, method='POST')
        task = self.create_task(uri, task_data)
        task.wait()

        return True

    def db(self, name):
        """Fetch a database associated with the environment.

        Parameters
        ----------
        name : str
            The name of the database to retrieve.

        Returns
        -------
        Database
            The requested database resource object.
        """
        uri = ('%s/dbs/%s' % (self.uri, name))
        return Database(uri, self.auth)

    def dbs(self):
        """Fetch all databases associated with the environment.

        Returns
        -------
        DatabaseList
            Dictionary of the databases keyed by name.
        """
        dbs = DatabaseList(self.uri, self.auth)
        return dbs

    def deploy_code(self, path):
        """Deploy code to the environment.

        Parameters
        ----------
        path : string
            The git reference path (branch/tag name) to deploy.

        Returns
        -------
        bool
            Was the code successfully deployed?
        """
        uri = '{}/code-deploy'.format(self.uri)
        params = {'path': path}

        task_data = self.request(uri=uri, method='POST', params=params)
        task = self.create_task(uri, task_data)
        task.wait()

        return True

    def domain(self, name):
        """Fetch a domain resource object.

        Parameters
        ----------
        name : string
            The FQDN of the domain to lookup.

        Returns
        -------
        Domain
            The domain resource object.
        """
        uri = '{uri}/domains/{name}'.format(uri=self.uri, name=name)
        return Domain(uri, self.auth)

    def domains(self):
        """Fetch a list of domains associated with the environment.

        returns dict of domains keyed by FQDN.

        """
        domains = DomainList(self.uri, self.auth)
        return domains

    def livedev(self, enable, discard=True):
        """Enable or disable live dev for the domain.

        Parameters
        ----------
        enable : bool
             Enable live development for this environment?
        discard : bool
             Discard all non committed changes?

        Returns
        -------
        Environment
            This environment object.
        """
        action = 'enable'
        params = {}

        if not enable:
            action = 'disable'
        uri = '{uri}/livedev/{action}'.format(uri=self.uri, action=action)

        if discard:
            params['discard'] = 1

        task_data = self.request(uri=uri, method='POST', params=params)
        task = self.create_task(uri, task_data)
        task.wait()

        return self


    def server(self, hostname):
        """Fetch a server associated with the environment.

        name The hostname of the server to lookup.

        return Server object

        """
        uri = ('%s/servers/%s', (self.uri, hostname))
        return Server(uri, self.auth)

    def servers(self):
        """Fetch a list of servers associated with the environment.

        returns dict of servers keyed by hostname.

        """
        servers = ServerList(self.uri, self.auth)
        return servers



""" Dictionary of Acquia Cloud API domain resources. """

from .acquialist import AcquiaList
from .domain import Domain

class DomainList(AcquiaList):

    """ Dictionary of Acquia Cloud API domain resources. """

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(DomainList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def create(self, name):
        """Create a new domain object.

        name The FQDN to be created.

        return The new domain object.

        """
        uri = self.get_resource_uri(name)
        task_data = self.request(uri=uri, method='POST')
        task = self.create_task(uri, task_data)
        task.wait()

        domain = Domain(uri, self.auth)

        self.__setitem__(name, domain)

        return domain

    def fetch(self):
        """ Fetch and store domain objects. """
        domains = super(DomainList, self).request(uri=self.uri)
        for domain in domains:
            name = domain['name'].encode('ascii', 'ignore')
            domain_uri = self.get_resource_uri(name)
            self.__setitem__(name, Domain(domain_uri, self.auth, data=domain))

    def get_resource_uri(self, fqdn):
        """ Generate the domain resource URI.

        Parameters
        ----------
        fqdn : str
            The FDQN of the domain resource.

        Returns
        -------
        str
            The server URI.
        """
        return '{base_uri}/{fqdn}'.format(base_uri=self.uri, fqdn=fqdn)

    def set_base_uri(self, base_uri):
        """ Set the base URI for domain resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/domains'.format(base_uri)
        self.uri = uri



""" List of database backups. """

import json

from .acquialist import AcquiaList
from .backup import Backup

class BackupList(AcquiaList):

    """List of database backups for a site."""

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(BackupList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def create(self):
        """ Create a new backup."""
        task_data = self.request(method='POST')

        task = self.create_task(self.uri, task_data)
        task.wait()

        # For some reason Acquia encodes JSON as a string in a JSON object.
        result = json.loads(task['result'])
        backup_id = result['backupid']
        uri = '{uri}/{backup_id}'.format(uri=self.uri, backup_id=backup_id)
        backup = Backup(uri, self.auth)

        self.__setitem__(backup_id, backup)

        return backup

    def fetch(self):
        """ Fetch and store database object. """
        backups = super(BackupList, self).request(uri=self.uri)
        for backup in backups:
            backup_id = backup['id'].encode('ascii', 'ignore')
            uri = self.get_resource_uri(backup_id)
            self.__setitem__(backup_id, Backup(uri, self.auth, data=backup))

    def set_base_uri(self, base_uri):
        """ Set the base URI for backup resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/backups'.format(base_uri)
        self.uri = uri



""" Dictionary of Acquia Cloud API site resources. """

from .acquialist import AcquiaList
from .site import Site

class SiteList(AcquiaList):

    """ Dictionary of site resources. """

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(SiteList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def fetch(self):
        """ Fetch and store site objects. """
        sites = super(SiteList, self).request(uri=self.uri)
        for site in sites:
            realm, name = site.encode('ascii', 'ignore').split(':')
            site_uri = self.get_resource_uri(name, realm=realm)
            self.__setitem__(name, Site(site_uri, self.auth))

    def get_resource_uri(self, name, realm='prod'):
        """ Generate the server URI.

        Parameters
        ----------
        name : str
            The subscription name of the site.

        Returns
        -------
        str
            The site URI.
        """
        return '{base_uri}/{realm}:{name}'.format(base_uri=self.uri, realm=realm, name=name)

    def set_base_uri(self, base_uri):
        """ Set the base URI for site resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/sites'.format(base_uri)
        self.uri = uri



""" Acquia Cloud API site resource. """

from .acquiaresource import AcquiaResource
from .environment import Environment
from .environmentlist import EnvironmentList
from .tasklist import TaskList
from .task import Task

class Site(AcquiaResource):

    """Site (or subscription) resource."""

    #: Valid keys for site object.
    valid_keys = ['title', 'name', 'production_mode', 'unix_username', 'vcs_type', 'vcs_url']

    def copy_code(self, source, target):
        """ Copy code from one environment to another.

        Parameters
        ----------
        source : str
            The name of the source environment.
        target : str
            The name of the target environment.

        Returns
        -------
        bool
            Was the code successfully copied?
        """
        uri = '{base}/code-deploy/{source}/{target}'.format(
            base=self.uri,
            source=source, target=target)

        task_data = self.request(uri=uri, method='POST')
        task = self.create_task(uri, task_data)
        task.wait()

        return True

    def environment(self, name):
        """ Retrieve an environment resource.

        Parameters
        ----------
        name : str
            The name of the environment.

        Returns
        -------
        Environment
            The environment resource object.
        """
        uri = '{uri}/envs/{name}'.format(uri=self.uri, name=name)
        return Environment(uri, self.auth)

    def environments(self):
        """ Retrieve a list of environments.

        Returns
        -------
        EnvironmentList
            Dictionary of environments keyed by name.
        """
        envs = EnvironmentList(self.uri, self.auth)
        return envs

    def task(self, task_id):
        """ Retrieve a task.

        Parameters
        ----------
        task_id : int
            The task identifier.

        Returns
        -------
        Task
            The task resource object.
        """
        uri = ('%s/tasks/%d' % (self.uri, task_id))
        return Task(uri, self.auth, hack_uri=False)

    def tasks(self):
        """ Retrieve all tasks for the site.

        Returns
        -------
        TaskList
            Dictionary of task resources keyed by id.
        """
        tasks = TaskList(self.uri, self.auth)
        return tasks



""" Imports all the resources. """

# Base classes
from .acquiadata import AcquiaData
from .acquialist import AcquiaList
from .acquiaresource import AcquiaResource

# Databases can be backed up
from .backup import Backup
from .backuplist import BackupList

# Environments have databases
from .database import Database
from .databaselist import DatabaseList

# Environments have domains
from .domain import Domain
from .domainlist import DomainList

# Sites reference environments
from .environment import Environment
from .environmentlist import EnvironmentList

# Sites have servers
from .server import Server
from .serverlist import ServerList

# Sites have tasks
from .task import Task
from .tasklist import TaskList

# Site is our top level
from .site import Site
from .sitelist import SiteList

# Users are top level
from .user import User



""" Acquia Cloud API domain resource. """

import re
from .acquiaresource import AcquiaResource

class Domain(AcquiaResource):

    """ Domain record associated with an environment. """

    valid_keys = ['name']

    def cache_purge(self):
        """Purge the varnish cache for the domain.

        Returns
        -------
        bool
            Was the cache cleared?
        """
        uri = '{}/cache'.format(self.uri)
        data = self.request(uri=uri, method='DELETE')
        task = self.create_task(uri, data)
        task.wait()

        return True

    def delete(self):
        """Delete the domain record.

        Returns
        -------
        bool
            Was the domain record deleted?
        """
        data = self.request(method='DELETE')
        task = self.create_task(self.uri, data)
        task.wait()
        return True

    def move(self, target):
        """Move a domain from one environment to another.

        Parameters
        ----------
        target : str
            The name of the environment the domain is being moved to.

        Returns
        -------
        Domain
            The new domain object or None is the move fails.
        """
        # These regex hacks are needed because Acquia doesn't keep this
        # function with domains, which sucks.
        pattern = re.compile('/envs/(.*)/domains/(.*)')
        matches = pattern.search(self.uri)
        current_env = matches.group(1)
        domain = matches.group(2)

        move_uri = ('%s/%s' % (pattern.sub(r'/domain-move/\g<1>', self.uri), target))
        data = {'domains': [domain]}

        task_data = self.request(uri=move_uri, method='POST', data=data)
        task = self.create_task(move_uri, task_data)
        task.wait()

        # Another hack, this time to get the URI for the domain.
        env_search = '/{}/'.format(current_env)
        env_target = '/{}/'.format(target)
        new_uri = self.uri.replace(env_search, env_target)
        return Domain(new_uri, self.auth)



""" Database resource. """

import re
from .acquiaresource import AcquiaResource
from .backup import Backup
from .backuplist import BackupList

class Database(AcquiaResource):

    """Environment database."""

    #: Valid keys for environment object.
    valid_keys = ['db_cluster', 'host', 'instance_name', 'name', 'password', 'username']

    def backup(self, backup_id):
        """Fetch a backup.

        Parameters
        ----------
        backup_id : int
            The id of the database backup to fetch.

        Returns
        -------
        Backup
            The backup object.
        """
        uri = '{uri}/backups/{backup_id}'.format(uri=self.uri, backup_id=backup_id)
        return Backup(uri, self.auth)

    def backups(self):
        """Fetch a list of database backups."""
        backups = BackupList(self.uri, self.auth)
        return backups

    def copy(self, target):
        """Copy a database to another environment.

        Parameters
        ----------
        target : string
            Name of the target environment.

        Returns
        -------
        Database
            Target database object.
        """
        # More regex hacks to work around the limitations of the ACAPI.
        pattern = re.compile('/envs/(.*)/dbs/(.*)')
        matches = pattern.search(self.uri)
        current_env = matches.group(1)
        base_uri = pattern.sub(r'/dbs/\g<2>/db-copy/\g<1>', self.uri)

        copy_uri = '{uri}/{target}'.format(uri=base_uri, target=target)

        task_data = self.request(uri=copy_uri, method='POST')
        task = self.create_task(copy_uri, task_data)
        task.wait()

        # Another hack, this time to get the URI for the domain.
        env_search = '/{}/'.format(current_env)
        env_target = '/{}/'.format(target)
        new_uri = self.uri.replace(env_search, env_target)
        return Database(new_uri, self.auth)



""" Acquia Cloud API user resource. """

from .acquiaresource import AcquiaResource

class User(AcquiaResource):

    """ Acquia Cloud API user resource. """

    #: Valid keys for a user.
    valid_keys = ['authenticated_as']

    def drushrc(self):
        """ Fetch all the drush aliases for sites the user has access to.

        The json+PHP output of this isn't very useful in python.

        Returns
        -------
        dict
            Collection of PHP code snippets.
        """
        uri = '{}/drushrc'.format(self.uri)
        response = self.request(uri=uri)
        return response



""" Dictionary of Acquia Cloud API task resources. """

from .acquialist import AcquiaList
from .task import Task

class TaskList(AcquiaList):

    """ Dictionary of task resources. """

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(TaskList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def fetch(self):
        """ Fetch and store task objects. """
        tasks = super(TaskList, self).request(uri=self.uri)
        for task in tasks:
            task_id = int(task['id'])
            task_uri = self.get_resource_uri(task_id)
            self.__setitem__(task_id, Task(task_uri, self.auth, data=task))

    def set_base_uri(self, base_uri):
        """ Set the base URI for task resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/tasks'.format(base_uri)
        self.uri = uri



""" Dictionary of Acquia Cloud API resources. """

from .acquiadata import AcquiaData
from ..exceptions import AcquiaCloudNoDataException

class AcquiaList(AcquiaData, dict):

    """ Dictionary of Acquia Cloud API resources. """

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ See :func:AcquiaData.__init__. """
        self.sorted_keys = []

        # Setup the dict
        dict.__init__(self, *args, **kwargs)

        # Setup the list
        self.set_base_uri(base_uri)
        AcquiaData.__init__(self, self.uri, auth)

    def __delitem__(self, key):
        """ See dict.__del_item__(). """
        super(AcquiaList, self).__delitem__(key)
        self.sorted_keys = []

    def __setitem__(self, key, value):
        """ See dict.__set_item__(). """
        super(AcquiaList, self).__setitem__(key, value)
        self.sorted_keys = []

    def get_resource_uri(self, res):
        """ Generate the resource URI.

        Parameters
        ----------
        res : mixed
            The identifier of the resource.

        Returns
        -------
        str
            The resource URI.
        """
        return '{base_uri}/{res}'.format(base_uri=self.uri, res=res)

    def first(self):
        """ Retrieve the first item in the dictionary.

        Returns
        -------
        AcquiaResource
            The first resource in the dictionary.
        """
        if not len(self):
            raise AcquiaCloudNoDataException('No data available')

        key = self.search_pos(0)
        return self[key]

    def get_sorted_keys(self):
        """ Get a sorted copy of the dictionary keys. """
        if not self.sorted_keys:
            keys = self.keys()
            self.sorted_keys = sorted(keys)

        return self.sorted_keys

    def last(self):
        """ Retrieve the last item in the dictionary.

        Returns
        -------
        AcquiaResource
            The last resource in the dictionary.
        """
        if not len(self):
            raise AcquiaCloudNoDataException('No data available')

        key = self.search_pos(-1)
        return self[key]

    def search_pos(self, pos):
        """ Search for a particular key.

        Parameters
        ----------
        pos : int
            The position of the key to return

        Returns
        -------
        str
            The key that occupies that position.
        """
        keys = self.get_sorted_keys()
        return keys[pos]

    def set_base_uri(self, base_uri):
        """ Set the base URI for the resource.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        self.uri = base_uri



""" Acquia Cloud API data resource. """

import json
import logging
import requests
import requests_cache
import time

from platform import python_version
from pprint import pformat
from ..version import __version__

LOGGER = logging.getLogger('acapi.resources.acquiadata')

class AcquiaData(object):

    """Acquia Cloud API abstract network resource."""

    #: User Agent string
    USER_AGENT = 'Acquia Cloud API Client/{mver} (Python {pver})'.format(mver=__version__,
                                                                         pver=python_version())

    def __init__(self, uri, auth, data=None):
        """ Constructor.

        Parameters
        ----------
        uri : str
            The base URI for the resource.
        auth : tuple
             The authentication credentials to use for the request.
        data : dict
            Raw data from ACAPI.
        """
        self.uri = uri
        self.auth = auth
        self.data = data
        self.last_response = None

    def create_task(self, uri, data):
        """ Create a new task object from a responses response object.

        Parameters
        ----------
        uri: str
            The URI for the action that triggered the task.
        data: dict
            The task data returned by the triggering request.

        Returns
        -------
        Task
            The Task object.
        """
        # We have to do this here to avoid circular dependencies
        from .task import Task
        task = Task(uri, self.auth, data=data)
        return task

    def get_last_response(self):
        """ Fetch the last response object. """
        return self.last_response

    def request(self, uri=None, method='GET', data=None, params={}, decode_json=True):
        """Perform a HTTP requests.

        Parameters
        ----------
        uri : str
            The URI to use for the request.
        method : str
            The HTTP method to use for the request.
        auth : tuple
            The authentication credentials to use for the request.
        data : dict
            Any data to send as part of a post request body.
        params : dict
            Query string parameters.

        Returns
        -------
        dict
            Decoded JSON response data as a dict object.
        """
        self.last_response = None

        if None == uri:
            uri = self.uri

        headers = {'User-Agent': self.USER_AGENT}

        uri = '{}.json'.format(uri)

        if 'GET' == method:
            attempt = 0
            while attempt <= 5:
                resp = requests.get(uri, auth=self.auth, headers=headers, params=params)

                if resp.status_code not in range(500, 505):
                    # No need to retry for if not a server error type.
                    break

                attempt += 1
                params['acapi_retry'] = attempt
                time.sleep((attempt ** 2.0) / 10)


            # We need to unset the property or it sticks around.
            if 'acapi_retry' in params:
                del params['acapi_retry']

        if 'POST' == method:
            jdata = json.dumps(data)
            resp = requests.post(uri, auth=self.auth, headers=headers, params=params, data=jdata)
            # This is a sledgehammer but fine grained invalidation is messy.
            requests_cache.clear()

        if 'DELETE' == method:
            resp = requests.delete(uri, auth=self.auth, headers=headers, params=params)
            # Quickest and easiest way to do this.
            requests_cache.clear()

        if hasattr(resp, 'from_cache') and resp.from_cache:
            LOGGER.info("%s %s returned from cache", method, uri)

        self.last_response = resp

        if resp.status_code != requests.codes.ok:
            try:
                raise resp.raise_for_status()
            except requests.exceptions.HTTPError as exp:
                LOGGER.info("Failed request response headers: \n%s",
                            pformat(exp.response.headers, indent=2))
                raise

        if decode_json:
            return resp.json()

        return resp.content



""" Acquia Cloud API Exceptions. """

from pprint import pformat

class AcquiaCloudException(Exception):

    """  Generic Acquia Cloud API Exception.

    All ACAPI exceptions should extend this class.
    """

    pass

class AcquiaCloudNoDataException(AcquiaCloudException):

    """ No data found exception. """

    pass

class AcquiaCloudTaskFailedException(AcquiaCloudException):

    """An Acquia task failure exception."""

    def __init__(self, message, task):
        """ Constructor.

        Parameters
        ----------
        message: str
            The error message.
        task: Task
            The Task object for the task that failed.
        """
        super(AcquiaCloudTaskFailedException, self).__init__(message)
        self.message = message
        self.task = task

    def __str__(self):
        """ Return the string representation of the exception.

        Returns
        -------
        str
            The error message and pretty printed Task object properties.
        """
        return "{msg}\n{task}".format(msg=self.message, task=pformat(self.task, indent=4))



""" Acquia Cloud API server list resource. """

from .acquialist import AcquiaList
from .environment import Environment

class EnvironmentList(AcquiaList):

    """Dict of Acquia Cloud API Environment resources keyed by short name."""

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(EnvironmentList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def fetch(self):
        """ Fetch and store environment objects. """
        envs = super(EnvironmentList, self).request(uri=self.uri)
        for env in envs:
            name = env['name'].encode('ascii', 'ignore')
            env_uri = self.get_resource_uri(name)
            self.__setitem__(name, Environment(env_uri, self.auth, data=env))

    def get_resource_uri(self, name):
        """ Generate the resource URI.

        Parameters
        ----------
        name : str
            The name of the environment resource.

        Returns
        -------
        str
            The resource URI.
        """
        return '{base_uri}/{name}'.format(base_uri=self.uri, name=name)

    def set_base_uri(self, base_uri):
        """ Set the base URI for server resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/envs'.format(base_uri)
        self.uri = uri



#!/usr/bin/env python

"""Setup ACAPI package."""

from setuptools import setup

setup(
    name='acapi',
    version='0.4.1',
    description='Acquia Cloud API client.',
    author='Dave Hall',
    author_email='me@davehall.com.au',
    url='http://github.com/skwashd/python-acquia-cloud',
    install_requires=['requests>=2.7', 'requests-cache==0.4.10'],
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Internet',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        ],
    packages=[
        'acapi',
        'acapi.resources',
        ],
    )



""" Acquia Cloud API database list resource. """

import re
from .acquialist import AcquiaList
from .database import Database

class DatabaseList(AcquiaList):

    """Dictionary of Acquia Cloud API database resources keyed by name."""

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(DatabaseList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def create(self, name):
        """ Create a new database.

        Parameters
        ----------
        name: str
            The name of the new database.

        Returns
        -------
        Database
            The new database object.
        """
        base_uri = re.sub(r'/envs/(.+)/dbs', '', self.uri)
        uri = '{base_uri}/dbs'.format(base_uri=base_uri)

        task_data = self.request(method='POST', uri=uri, data={'db': name})
        task = self.create_task(uri, task_data)
        task.wait()

        db_uri = '{uri}/{name}'.format(uri=self.uri, name=name)
        db_obj = Database(db_uri, self.auth)

        self.__setitem__(name, db_obj)

        return db_obj

    def fetch(self):
        """ Fetch and store database objects. """
        dbs = super(DatabaseList, self).request(uri=self.uri)
        for db_obj in dbs:
            name = db_obj['name'].encode('ascii', 'ignore')
            db_uri = self.get_resource_uri(name)
            self.__setitem__(name, Database(db_uri, self.auth, data=db_obj))

    def get_resource_uri(self, name):
        """ Generate the database URI.

        Parameters
        ----------
        name : str
            The name of the database.

        Returns
        -------
        str
            The database URI.
        """
        return '{base_uri}/{name}'.format(base_uri=self.uri, name=name)

    def set_base_uri(self, base_uri):
        """ Set the base URI for database resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/dbs'.format(base_uri)
        self.uri = uri



""" Acquia Cloud API server list resource. """

from .acquialist import AcquiaList
from .server import Server

class ServerList(AcquiaList):

    """Dict of Acquia Cloud API Server resources keyed by hostname."""

    def __init__(self, base_uri, auth, *args, **kwargs):
        """ Constructor. """
        super(ServerList, self).__init__(base_uri, auth, *args, **kwargs)
        self.fetch()

    def fetch(self):
        """ Fetch and store server objects. """
        servers = super(ServerList, self).request(uri=self.uri)
        for server in servers:
            name = server['name'].encode('ascii', 'ignore')
            server_uri = self.get_resource_uri(name)
            self.__setitem__(name, Server(server_uri, self.auth, data=server))

    def get_resource_uri(self, name):
        """ Generate the server URI.

        Parameters
        ----------
        name : str
            The hostname of the server.

        Returns
        -------
        str
            The server URI.
        """
        return '{base_uri}/{name}'.format(base_uri=self.uri, name=name)

    def set_base_uri(self, base_uri):
        """ Set the base URI for server resources.

        Parameters
        ----------
        base_uri : str
            The base URI to use for generating the new URI.
        """
        uri = '{}/servers'.format(base_uri)
        self.uri = uri



"""Database Backup."""

import urlparse

from .acquiaresource import AcquiaResource

class Backup(AcquiaResource):

    """Acquia Cloud database backup."""

    valid_keys = ['completed', 'created', 'description', 'id', 'percentage', 'queue',
                  'recipient', 'result', 'sender', 'started', 'state']

    def delete(self):
        """Delete the current backup resource.

        Returns
        -------
        bool
            Was the backup deleted?

        """
        self.request(method='DELETE')
        return True

    def download(self, target_file):
        """Download a database backup from Acquia Cloud & write it to a file.

        Parameters
        ----------
        target_file : str
            The file to use for the backup. Overwrites file if it exists.

        Returns
        -------
        bool
            Was the backup retrieved and saved?

        """
        response = self.request()
        backup = response

        # Hack to make the download URL work for requests
        url_parts = urlparse.urlparse(backup['link'])
        query = urlparse.parse_qs(url_parts.query)
        url_parts._replace(query='')
        uri = url_parts.geturl()

        content = self.request(uri=uri, params=query, decode_json=False)

        backup_file = open(target_file, 'wb')
        backup_file.write(content)
        backup_file.close()

        return True



"""Acquia Cloud API client."""

import os
import requests_cache

from .resources import Site, SiteList, User
from .exceptions import AcquiaCloudException

class Client(object):

    """A Client for accessing the Acquia Cloud API."""

    def __init__(
            self, user=None, token=None, realm='prod',
            endpoint='https://cloudapi.acquia.com/v1',
            cache=600):
        """Create an Acquia Cloud API REST client.

        Parameters
        ----------
        user : str
            Acquia Cloud API username.
        token : str
            Acquia Cloud API user token.
        realm : str
            Acquia Cloud API realm (defaults to 'prod').
        endpoint : str
            Base Acquia Cloud API endpoint URL.
        cache: int
            How long API responses should be cached for.

        """
        if not user or not token:
            user, token = self.__find_credentials()
            if not user or not token:
                raise AcquiaCloudException("Credentials not provided")

        self.auth = (user, token)
        self.realm = realm
        self.endpoint = endpoint

        if None != cache:
            requests_cache.install_cache(cache_name='acapi', backend='memory', expire_after=cache)

    def generate_uri(self, path):
        """Generate a URI for a ACAPI request.

        Parameters
        ----------
        path : str
            The path component of the URI

        Returns
        -------
        str
            The generates URI.

        """
        uri = '{endpoint}/{path}'.format(endpoint=self.endpoint, path=path)
        return uri

    def site(self, name):
        """Retrieve a site object.

        Parameters
        ----------
        name : str
            The Acquia site/subscription name to look up.

        Returns
        -------
        Site
            The site object.

        """
        namespace = ('sites/%s:%s' % (self.realm, name))
        uri = self.generate_uri(namespace)
        site = Site(uri, self.auth)
        return site

    def sites(self):
        """Retrieve a list of available sites.

        Returns
        -------
        SiteList
            dict of sites keyed by site name.

        """
        sites = SiteList(self.endpoint, self.auth)
        return sites

    def user(self):
        """Retrieve the currently authenticated Cloud API user."""
        user = User(self.generate_uri('me'), self.auth)
        return user

    def __find_credentials(self):
        """Check environment variables for API credentials."""
        user = os.environ.get('ACQUIA_CLOUD_API_USER')
        token = os.environ.get('ACQUIA_CLOUD_API_TOKEN')
        return user, token
